#!/usr/bin/env python
##
## This file is part of Nolimips, a MIPS simulator with unlimited registers
## Copyright (C) 2003, 2004, 2006 Benoit Perrot <benoit@lrde.epita.fr>
##
## Nolimips is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## Nolimips is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
##

import sys, os
import string

import nolimips

#### Nolimips' instruction classes generator ---------------

if __name__ != "__main__" or len(sys.argv) != 2:
  print "Usage: .. SRCDIR"
  sys.exit(1)
srcdir = sys.argv[1]

## Generate the class corresponding to an instruction
format_type_to_enum = {
  "register": "r_type",
  "immediate": "i_type",
  "jump": "j_type",
  }

def class_generate(inst):

  ## Shortands
  class_id = nolimips.class_id(inst.opcode)
  define_id = nolimips.define_id(inst.opcode)
  file_id = nolimips.file_id(inst.opcode)
  banner = nolimips.license + "\n" + nolimips.generated_by_hand
  format = inst.format
  
  ## Open header, inline and implementation files
  header = file(file_id + ".hh.tmp", "w")
  inline = file(file_id + ".hxx.tmp", "w")
  impl = file(file_id + ".cc.tmp", "w")
  
  for f, ext, brief in [(header, "hh",  "Declare the interface"),
                        (inline, "hxx", "Inline methods"),
                        (impl,   "cc",  "Implementation")]:
    sys.stdout = f
    print banner
    # Document the file
    print "/** \\file inst/" + file_id + "." + ext
    print "    \\brief " + brief + " of " + class_id + " */"
    # Open multiple inclusion protection
    if ext != "cc":
      print "#ifndef INST_" + string.upper(inst.opcode + "_" + ext)
      print "# define INST_" + string.upper(inst.opcode + "_" + ext)

  sys.stdout = header
  print """
# include \"inst/register.hh\"
# include \"inst/exp.hh\"
# include \"inst/inst.hh\""""

  sys.stdout = impl
  print "#include \"inst/visitor.hh\""
  
  for f in [inline, impl]:
    sys.stdout = f
    print "#include \"inst/" + file_id + ".hh\""

  # Open namespace
  for f in [header, inline, impl]:
    sys.stdout = f
    print """
namespace inst
{
"""

  ## Declare class
  sys.stdout = header
  print "  /// Declare " + class_id + "."
  print "  /** " + inst.desc + "  */"
  print "  class " + class_id + ":"
  print "    public Inst"
  print "  {"
  
  # Constructor and destructor
  args = ""
  init = ""
  dels = ""
  for attribute in format.attributes:
    
    if attribute.type == "Register":
      args = args + "const " + attribute.type + " &"
    elif attribute.type == "Exp":
      args = args + "const CountedExp &"
    else:
      args = args + attribute.type + " *"
    args = args + attribute.name + ", "
    init = init + attribute.name + "_(" + attribute.name + "), "
  args = string.rstrip(args, ", ")
  init = string.rstrip(init, ", ")
  sys.stdout = header
  print """    /** \\name Ctor & dtor.
        \\{ */
  public:"""
  print "    /// Construct a(n) " + class_id + " node."
  print "    " + class_id + "(" + args + ");"
  print "    /// Destroy a(n) " + class_id + " node."
  print "    virtual ~" + class_id + "();"
  print "    /** \\} */"
  sys.stdout = inline
  print "  inline"
  print "  " + class_id + "::" + class_id + "(" + args + ")" + ":"
  if init == "":
    print "    Inst(" + format_type_to_enum[format.type] + ")"
  else:
    print "    Inst(" + format_type_to_enum[format.type] + "),"
  print "    " + init
  print "  {}\n"
  sys.stdout = impl
  print "  " + class_id + "::~" + class_id + "()"
  print "  {"
  for attribute in format.attributes:
    # FIXME: there should be some preconditions
    if attribute.type != "Register" and attribute.type != "Exp":
      print "    delete " + attribute.name + "_;"
  print "  }\n"

  # Methods: accessors
  sys.stdout = header
  print """
    /** \\name Accessors.
        \\{ */
  public:"""
  for attribute in format.attributes:
    for c in ["const ", ""]:
      sys.stdout = header
      print "    /// Return " + string.lower (attribute.name)
      print "    " + c + attribute.type + \
            " &\t\tget_" + attribute.name + "() " + c + ";"
      sys.stdout = inline
      print "  inline"
      print "  " + c + attribute.type + " &"
      print "  " + class_id + "::get_" + attribute.name + "() " + c
      print "  {"
      if attribute.type == "Register":
        print "    return " + attribute.name + "_;"
      else:
        print "    return *" + attribute.name + "_;"
      print "  }"
  sys.stdout = header
  print "    /** \\} */"
  
  # Method: visitor entry point
  sys.stdout = inline
  print ""
  sys.stdout = impl
  print "  void"
  print "  " + class_id + "::accept (ConstVisitor& v) const"
  print """  {
    v.visit(*this);
  }"""
  print "  void"
  print "  " + class_id + "::accept (Visitor& v)"
  print """  {
    v.visit(*this);
  }"""
  sys.stdout = header
  print """
    /** \\name Visitors entry point.
        \\{ */
  public:
    /// Accept a const visitor \\a v.
    virtual void accept(ConstVisitor& v) const;
    /// Accept a non-const visitor \\a v.
    virtual void accept(Visitor& v);
    /** \\} */
"""

  # Method: print
  sys.stdout = header
  print """  public:
    /// Print this
    virtual void\twrite(std::ostream &ostr) const;
"""
  sys.stdout = impl
  print """
  void"""
  print "  " + class_id + "::write(std::ostream &ostr) const"
  print "  {"
  s = ""
  i = 0
  while i < len(format.attributes):
    if (i < len(format.attributes) - 1) and \
           (format.attributes[i].type == "Exp") and \
           (format.attributes[i + 1].type == "Register"):
      s = s + " << *" + format.attributes[i].name + "_" + \
          " << '(' << " + format.attributes[i + 1].name + "_ << ')' "
      i = i + 1
    else:
      s = s + " << "
      if format.attributes[i].type != "Register":
        s = s + "*"
      s = s + format.attributes[i].name + "_ << \", \""
    i = i + 1
  s = string.rstrip(s, "<< \", \"")
  print "    ostr << \"" + inst.opcode + "\\t\"" + s + ";"
  print "  }"

  # Declare attributes
  sys.stdout = header
  if len(format.attributes) > 0:
    print "  protected:"
    for attribute in format.attributes:
      id = attribute.name + "_"
      if attribute.type == "Register":
        print "    " + attribute.type + "\t" + id + ";"
      elif attribute.type == "Exp":
        print "    CountedExp\t" + id + ";"
      else:
        print "    " + attribute.type + "\t*" + id + ";"

  # End of declaration
  print "  };"


  ## Close files
  for f, ext in [(header, "hh"), (inline, "hxx"), (impl, "cc")]:
    # Close namespace and inclusion protection system
    sys.stdout = f
    print "\n} // namespace inst"
    if ext != "cc":
      if ext == "hh":
        print "\n# include \"inst/" + file_id + ".hxx\""
      print "\n#endif // !INST_" + string.upper(inst.opcode + "_" + ext)
    # Check differences between old and new files
    sys.stdout = sys.__stdout__
    f.close()
    nolimips.lazy_overwrite(srcdir + os.sep + file_id + "." + ext,
                            file_id + "." + ext + ".tmp")
    os.remove(file_id + "." + ext + ".tmp")

## Main --------------------------------
parser = nolimips.InstructionSetParser()
instructions = parser.parse(sys.stdin)

## Generate classes files
for inst in instructions:
  if inst.level == "native":
    class_generate(inst)

## Generate all-includer
sys.stdout = file("all.hh.tmp", "w")
print nolimips.license
print nolimips.generated_by_hand
print "#ifndef INST_ALL_HH"
print "# define INST_ALL_HH"
for inst in instructions:
  if inst.level == "native":
    print "#  include \"inst/" + inst.opcode + ".hh\""
print "#endif // !INST_ALL_HH"
sys.stdout.close()
sys.stdout = sys.__stdout__
nolimips.lazy_overwrite(srcdir + os.sep + "all.hh", "all.hh.tmp")
os.remove("all.hh.tmp")

## Generate all-declarator
sys.stdout = file("fwd.hh.tmp", "w")
print nolimips.license
print nolimips.generated_by_hand
print """
#ifndef INST_FWD_HH
# define INST_FWD_HH

# include \"misc/select_const.hh\"

namespace inst
{

  class Register;
  class Label;
  class TextLabel;

  class Exp;
  class IntExp;
  class LabelExp;
  class OpExp;

  template <typename ConstKind>
  class GenExpVisitor;
  typedef GenExpVisitor<const_kind> ConstExpVisitor;
  typedef GenExpVisitor<non_const_kind> ExpVisitor;

  class Inst;
"""
for inst in instructions:
  if inst.level == "native":
    print "  class " + nolimips.class_id(inst.opcode) + ";"
print """

  class Section;
  class TextSection;
  class DataSection;

  class Program;
  class ProgramBuilder;
  class ProgramSolver;

} // namespace inst

#endif // !INST_FWD_HH
"""
sys.stdout.close()
sys.stdout = sys.__stdout__
nolimips.lazy_overwrite(srcdir + os.sep + "fwd.hh", "fwd.hh.tmp")
os.remove("fwd.hh.tmp")

## Generate visitor
sys.stdout = file("visitor.hh.tmp", "w")
print nolimips.license
print nolimips.generated_by_hand
print """#ifndef INST_VISITOR_HH
# define INST_VISITOR_HH

# include \"misc/select_const.hh\"

# include \"inst/fwd.hh\"

namespace inst
{

  template <typename K>
  class GenVisitor
  {
  public:
    virtual ~GenVisitor()
    {
    }

  public:
  """
for inst in instructions:
  if inst.level == "native":
    print "    virtual void\tvisit(typename select_const<K, " \
          + nolimips.class_id(inst.opcode) + ">::type&) = 0;"
print """
  };

  typedef GenVisitor<const_kind> ConstVisitor;
  typedef GenVisitor<non_const_kind> Visitor;

} // namespace inst

#endif // !INST_VISITOR_HH"""
sys.stdout.close()
sys.stdout = sys.__stdout__
nolimips.lazy_overwrite(srcdir + os.sep + "visitor.hh", "visitor.hh.tmp")
os.remove("visitor.hh.tmp")
