//
// This file is part of Nolimips, a MIPS simulator with unlimited registers
// Copyright (C) 2003, 2004, 2006 Benoit Perrot <benoit@lrde.epita.fr>
//
// Nolimips is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// Nolimips is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
//
/** \file shell/shell.hh
    \brief Declare Shell class */

#ifndef SHELL_SHELL_HH
# define SHELL_SHELL_HH

# include <list>
# include <map>
# include <string>
# include <sstream>
# include <iostream>

# include "vm/virtual_machine.hh"
# include "shell/cmd.hh"

namespace shell
{

  /// Shell Class
  class Shell
  {
  public:
    /** \name Shortand type definition
	\{ */
    typedef std::map<std::string, Cmd::identifier_type> map_token_type;
    typedef std::map<std::string, vm::Cpu::kind_type> map_register_type;
    typedef std::list<std::string> list_disp_type;
    /** \} */

    /** \name Constructor and destructor
	\{ */
  protected:
    /// Construct a Shell
    Shell();

  public:
    static Shell &instance();

    /// Destroy a Shell
    ~Shell();
    /** \} */

  public:
    /// Run the command interpreter
    void run();

    /// Build a command
    Cmd *parse(const std::string &str,
	       std::vector<std::string> *completions = 0);

  public:
    /// Access a register
    int get_register(const std::string &r);

    /** \name Display list
	\{ */
  public:
    /// Add a register in the display list
    void add_display(const std::string &r);

    /// Remove a register from the display list
    void remove_display(const std::string &r);

    /// Display the register in the diplay list
    void display();
    /** \} */

    // Internal scanner/parser
  private:
    /// Eat the next word from input stringstream
    std::string eat_word(std::istringstream &iss);

    /// Eat a valid register identifier from input stringstream
    bool eat_register(std::istringstream &iss, std::string &reg);

    /// Eat a valid label from input stringstream
    bool eat_label(std::istringstream &iss, std::string &label);

    /// Eat a word from input stringstream and return the
    /// corresponding keyword
    Cmd::identifier_type eat_command_id(std::istringstream &iss,
					std::vector<std::string> *completions);

    /** \name Execution engine
	\{ */
  protected:
    void do_load(const Cmd &cmd);
    void do_run();
    void do_continue();
    void do_next();
    void do_break(const Cmd &cmd);

    void do_display(const Cmd &cmd);
    void do_undisplay(const Cmd &cmd);
    void do_dump();
    void do_print(const Cmd &cmd);
    void do_mem();
    void do_backtrace() const;

    void do_nop_after_branch();

    void do_unknown() const;
    void do_ambiguous() const;

    void do_help() const;

    int execute(const Cmd &cmd);
    /** \} */

  private:
    /// Program being run
    inst::Program *program_;

    /// Virtual machine
    vm::VirtualMachine vm_;

    /// map of token word
    map_token_type map_token_;
    /// map of register
    map_register_type map_register_;
    /// list of display register
    list_disp_type list_displayable_;
  }; // class Shell

} // namespace shell

# include "shell/shell.hxx"

#endif // !SHELL_SHELL_HH
